package jpopup.contact;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.HashMap;
import java.util.Iterator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class used to manipulate contact notes, load them from xml, save them to xml, add and delete contacts
 * @author Alexey Kalukin
 *
 */
public class ContactList implements Iterable<Contact> {
	private static final String ROOT_TAG = "contacts";

	private static final String CONTACT_TAG = "contact";

	private static final String TITLE_TAG = "title";
	
	private static final String UID_TAG = "uid";

	private static final String FIRST_NAME_TAG = "firstName";

	private static final String LAST_NAME_TAG = "lastName";

	private static final String EMAIL_TAG = "e-mail";

	private static final String ICQ_TAG = "icq";

	private static final String ADDITIONAL_INFO_TAG = "additionalInfo";

	private HashMap<String, Contact> contactList = new HashMap<String, Contact>();

	DocumentBuilderFactory dbf;

	DocumentBuilder parser;

	public ContactList() throws ParserConfigurationException {
		dbf = DocumentBuilderFactory.newInstance();
		dbf.setValidating(false);

		parser = dbf.newDocumentBuilder();

		parser.setErrorHandler(new ErrorHandler() {

			public void error(SAXParseException exception) throws SAXException {
				System.err.println("ERROR: " + exception.getMessage());
			}

			public void fatalError(SAXParseException exception)
					throws SAXException {
				System.err.println("FATAL: " + exception.getMessage());
				throw exception;
			}

			public void warning(SAXParseException exception)
					throws SAXException {
				System.err.println("WARNING: " + exception.getMessage());
			}

		});
	}

	public ContactList(File contactFile) throws SAXException, IOException,
			ParserConfigurationException, ParseException {
		this();
		getContactsFromXMLFile(contactFile);
	}

	public Iterator<Contact> iterator() {
		return contactList.values().iterator();
	}

	public void getContactsFromXMLFile(File xmlFile) throws SAXException,
			IOException, ParserConfigurationException, ParseException {
		Document xml = parser.parse(xmlFile);
		getContactsFromXML(xml);
	}

	public void putContactsToXMLFile(String fileName) throws IOException,
			TransformerException {
		Document xml = putContactsToXML();

		TransformerFactory factory = TransformerFactory.newInstance();
		Transformer transformer = factory.newTransformer();
		transformer.transform(new DOMSource(xml), new StreamResult(new File(
				fileName)));		
	}

	public void addContact(Contact contact) {
		contactList.put(contact.getUID(), contact);
	}

	public Contact getContact(String title) {
		return contactList.get(title);
	}

	public void removeContact(String title) {
		contactList.remove(title);
	}

	private void getContactsFromXML(Document xml) throws ParseException {
		NodeList contacts = xml.getElementsByTagName(CONTACT_TAG);

		int contactsCount = contacts.getLength();
		for (int i = 0; i < contactsCount; i++) {
			Element contact = (Element) contacts.item(i);
			Contact newContact = new Contact(getXMLField(contact, UID_TAG));
			newContact.setFirstName(getXMLField(contact, FIRST_NAME_TAG));
			newContact.setTitle(getXMLField(contact, TITLE_TAG));
			newContact.setLastName(getXMLField(contact, LAST_NAME_TAG));
			newContact.setEmail(getXMLField(contact, EMAIL_TAG));
			newContact.setICQ(getXMLField(contact, ICQ_TAG));
			newContact.setAdditionalInfo(getXMLField(contact,
					ADDITIONAL_INFO_TAG));

			contactList.put(newContact.getUID(), newContact);
		}
	}

	private Document putContactsToXML() {
		Document xml = parser.newDocument();
		xml.appendChild(xml.createElement(ROOT_TAG));
		Node rootNode = xml.getElementsByTagName(ROOT_TAG).item(0);

		for (Contact contact : contactList.values()) {
			putContact2XML(rootNode, contact);
		}

		return xml;
	}

	private void putContact2XML(Node parent, Contact contact) {
		Node contactNode = parent.getOwnerDocument().createElement(CONTACT_TAG);
		putField2XML(contactNode, contact.getUID(), UID_TAG);
		putField2XML(contactNode, contact.getTitle(), TITLE_TAG);
		putField2XML(contactNode, contact.getFirstName(), FIRST_NAME_TAG);
		putField2XML(contactNode, contact.getLastName(), LAST_NAME_TAG);
		putField2XML(contactNode, contact.getEmail(), EMAIL_TAG);
		putField2XML(contactNode, contact.getICQ(), ICQ_TAG);
		putField2XML(contactNode, contact.getAdditionalInfo(),
				ADDITIONAL_INFO_TAG);

		parent.appendChild(contactNode);
	}

	private void putField2XML(Node parent, String data, String fieldName) {
		Node node = parent.getOwnerDocument().createElement(fieldName);
		node.appendChild(parent.getOwnerDocument().createTextNode(data));
		parent.appendChild(node);
	}

	private String getXMLField(Element element, String fieldName) {
		String fieldData = "";
		NodeList dataList = element.getElementsByTagName(fieldName);
		Node dataField = dataList.item(0).getFirstChild();
		if (dataField != null) {
			fieldData = ((Text) dataField).getData().trim();
		}

		return fieldData;
	}

}
