package jpopup.history;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import javax.swing.JOptionPane;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class used for manipulating history messages, save/load messages to XML file
 * @author Alexey Kalukin
 *
 */
public class XMLMessageHistory extends MessageHistory {
	private static final String ROOT_TAG = "history";
	private static final String MSG_TAG = "msg";
	private static final String SENDER_TAG = "sender";
	private static final String TEXT_TAG = "text";
	private static final String TIME_TAG = "time";

	private String xmlFileName;
	private BufferedWriter historyWriter;

	private ArrayList<Message> history = new ArrayList<Message>();
	private DocumentBuilderFactory dbf;

	private DocumentBuilder parser;
	private  static LinkedHashMap<String, String> escapeXmls = new LinkedHashMap<String, String>();

	public XMLMessageHistory(String contactTitle)
			throws ParserConfigurationException, IOException {
		super(contactTitle);
		String title = new String(contactTitle);
		if (title.equals(MessageHistory.DEFAULT_GROUP)) {
			title = MessageHistory.DEFAULT_GROUP_RENAMER;
		}
		File historyDir = new File(HISTORY_DIRECTORY);
		if (!historyDir.exists()) {
			historyDir.mkdir();
		}
		xmlFileName = MessageHistory.HISTORY_DIRECTORY + "/" + title
				+ "." + MessageHistory.HISTORY_FILE_EXT;		
		
		historyWriter = new BufferedWriter(new FileWriter(xmlFileName, true));
		
		dbf = DocumentBuilderFactory.newInstance();
		dbf.setValidating(false);

		parser = dbf.newDocumentBuilder();
		parser.setErrorHandler(new ErrorHandler() {

			public void error(SAXParseException exception) throws SAXException {
				System.err.println("ERROR: " + exception.getMessage());
			}

			public void fatalError(SAXParseException exception)
					throws SAXException {
				System.err.println("FATAL: " + exception.getMessage());
				throw exception;
			}

			public void warning(SAXParseException exception)
					throws SAXException {
				System.err.println("WARNING: " + exception.getMessage());
			}

		});
		
		checkXMLFileExist();
		
		getEscapeXmls();
	}
	
	public void addMessage(String sender, String msg, String timeStamp) {
		try {
			Message newMessage = new Message(sender, msg, timeStamp);
			putMessage2XML(newMessage);
			
		} catch (ParseException e) {			
			// e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Can't write new message to the history. Parse error: wrong string time format.\n" + e, "Error", JOptionPane.ERROR_MESSAGE);
		} catch (IOException e) {
			// e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Can't write new message to the history. IO Error with history xml file " + xmlFileName + ".\n" + e, "Error", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	public void removeMessage(Message msg) {
		history.remove(msg);		
	}


	public void getHistoryFromXMLFile() throws SAXException, IOException, ParseException {
		FakeInputStream ffis = new FakeInputStream(xmlFileName);	
		Document xml = parser.parse(ffis);

		NodeList msgs = xml.getElementsByTagName(MSG_TAG);

		int msgsCount = msgs.getLength();
		String sender;
		String text;
		String time;
		
		history = new ArrayList<Message>();
		for (int i = 0; i < msgsCount; i++) {
			Element msg = (Element) msgs.item(i);
			sender = getXMLField(msg, SENDER_TAG);
			text = getXMLField(msg, TEXT_TAG);
			time = getXMLField(msg, TIME_TAG);
			Message newMsg = new Message(sender, text, time);

			history.add(newMsg);
		}

	}
	
	public void putHistoryToXMLFile() throws IOException, TransformerException {
		File xmlFile = new File(xmlFileName);
		xmlFile.delete();
		xmlFile.createNewFile();
				
		putHistoryToXML();		
	}
	
	public Iterator<Message> iterator() {
		return history.iterator();
	}
	
	public ArrayList<Message> getLastMessages(int count) {
		ArrayList<Message> lastMSGS = new ArrayList<Message>();
		
		int startWith = history.size() - count; 
		for (int i = startWith >= 0 ? startWith : 0; i < history.size(); i++) {
			lastMSGS.add(history.get(i));
		}
		
		return lastMSGS;
	}
	
	private void createDefaultHistoryXMLFile() throws IOException, TransformerException {
		Document xml = parser.newDocument();
		xml.appendChild(xml.createElement(ROOT_TAG));		
		putHistoryToXMLFile();
	}
	
	private void checkXMLFileExist() {
		File file = new File(xmlFileName);
		if (!file.exists()) {
			try {
				file.createNewFile();
				createDefaultHistoryXMLFile();
			} catch (IOException e) {
				// e.printStackTrace();
				JOptionPane.showMessageDialog(null, "Can't create default history file. IO Error with history xml file " + xmlFileName + ".\n" + e, "Error", JOptionPane.ERROR_MESSAGE);
			} catch (TransformerException e) {
				// e.printStackTrace();
				JOptionPane.showMessageDialog(null, "Can't create default history file. Error with parsing DOM tree to history xml file " + xmlFileName + ".\n" + e, "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
	}
	
	private String getXMLField(Element element, String fieldName) {
		String fieldData = "";
		NodeList dataList = element.getElementsByTagName(fieldName);
		Node dataField = dataList.item(0).getFirstChild();
		if (dataField != null) {
			fieldData = ((Text)dataField).getData().trim();
		}
		
		return fieldData;
	}
	
	private void putHistoryToXML() throws IOException {
		for (Message msg: history) {
			putMessage2XML(msg);			
		}		
	}
	
	private void putMessage2XML(Message msg) throws IOException {
		
		historyWriter.write("<" + MSG_TAG + ">");
		putField2XML(historyWriter , msg.getSender(), SENDER_TAG);
		putField2XML(historyWriter, msg.getText(), TEXT_TAG);
		putField2XML(historyWriter, msg.getFormatTime(), TIME_TAG);		
		historyWriter.write("</" + MSG_TAG + ">");
		
		historyWriter.flush();
	}
	
	private void putField2XML(BufferedWriter xmlFile, String data, String fieldName) throws IOException {
		xmlFile.write("<" + fieldName + ">");
		xmlFile.write(preparingTextMsg2Export(data));
		xmlFile.write("</" + fieldName + ">");
	}
	
	private String preparingTextMsg2Export(String dirtyText) {
		String clearMSG = new String(dirtyText);
		for (Entry<String, String> escapeXml : escapeXmls.entrySet()) {
			clearMSG = clearMSG.replaceAll(escapeXml.getKey(), escapeXml
					.getValue());
		}
		
		return clearMSG;
	}
	
	private void getEscapeXmls() {
		escapeXmls.put("[&]{1}", "&amp;");
		escapeXmls.put("[<]{1}", "&lt;");
		escapeXmls.put("[>]{1}", "&gt;");
		escapeXmls.put("[$]{1}", "\\\\$");
		escapeXmls.put("[%]{1}", "\\\\%");				
	}
}
